/* forward star type graph structure -- for quite large graphs */
/* 6/Nov/2008   Takeaki Uno  */

#ifndef _fstar_c_
#define _fstar_c_

#include"fstar.h"
#include"stdlib2.c"

FSTAR INIT_FSTAR = {TYPE_FSTAR,NULL,0,NULL,NULL,NULL,NULL,0,0,0,0,0,0,0,0,NULL,NULL,0, 0,0,0,FSTAR_INTHUGE,FSTAR_INTHUGE,FSTAR_INTHUGE};
QSORT_TYPE (FSTAR_INT,FSTAR_INT)

void FSTAR_print (FSTAR *F){
  FSTAR_INT i, x=0, y, flag;

  FLOOP (i, 0, F->edge_num+1){
    while ( i == F->fstar[x+1]){
      flag = 0;
      if ( F->out_deg && (F->out_node_num ==0 || x < F->out_node_num) ) flag += F->out_deg[x];
      if ( F->in_deg && (F->in_node_num ==0 || x < F->in_node_num) ) flag += F->in_deg[x];
      if ( (!F->out_deg && !F->in_deg) && (F->flag&LOAD_EDGE) ) flag = F->fstar[x+1] - F->fstar[x];
      if ( !F->table || flag ) printf ("\n");
      if ( ++x == F->out_node_num ) goto END;
    }
    y = F->edge[i];
    if ( !(F->flag&(LOAD_BIPARTITE+LOAD_EDGE)) && F->edge_dir==0 && x>y ) continue;
    if ( F->table && !(F->flag&LOAD_BIPARTITE) ){
      if ( (y=F->table[y]) == F->out_node_num ) continue;
    }
    if ( F->flag & LOAD_ID1 ) y++;
    printf (FSTAR_INTF" ", y);
  }
  printf ("\n");
  END:;
}

/* initialization */
void FSTAR_init2 (FSTAR *F){
  if ( F->in_lb >0 || F->in_ub <FSTAR_INTHUGE ) F->flag |= (FSTAR_IN_CHK+FSTAR_CNT_IN);
  if ( F->out_lb >0 || F->out_ub <FSTAR_INTHUGE ) F->flag |= (FSTAR_OUT_CHK+FSTAR_CNT_OUT);
  if ( F->deg_lb >0 || F->deg_ub <FSTAR_INTHUGE ){
    F->flag |= FSTAR_DEG_CHK+FSTAR_CNT_IN;
    if ( (F->flag&(FSTAR_OUT_CHK+FSTAR_IN_CHK)) ==0 ) F->flag |= FSTAR_CNT_DEG_ONLY;
    else F->flag |= FSTAR_CNT_OUT;
  }
}

/* termination */
void FSTAR_end (FSTAR *F){
  mfree (F->edge, F->in_deg, F->out_deg, F->fstar, F->table, F->rev_table);
  *F = INIT_FSTAR;
}

/* increment degrees */
void FSTAR_inc_deg (FSTAR *F, FSTAR_INT x, FSTAR_INT y){
  if ( F->out_deg ) F->out_deg[x]++;
  if ( F->out_deg && (F->flag&LOAD_EDGE) ) F->out_deg[y]++;
  if ( F->in_deg ) F->in_deg[y]++;
  if ( ((F->flag & FSTAR_CNT_DEG_ONLY) || F->edge_dir==0) && F->in_deg ) F->in_deg[x]++;
}

/* remove edges adjacent to small/large degree vertices */
void FSTAR_sort_adjacent_node (FSTAR *F){
  FSTAR_INT x;
  FLOOP (x, 0, F->out_node_num){
    qsort_FSTAR_INT (&F->edge[F->fstar[x]], F->fstar[x+1]-F->fstar[x], 
      (F->flag&LOAD_DECSORT)? -1: 1);
  }
}

/* read a node pair (edge) from file. return 1 if fails */
int FSTAR_read_pair (FSTAR *F, FILE2 *fp, FSTAR_INT *x, FSTAR_INT *y){
  *x = (FSTAR_INT)FILE2_read_int (fp);
  if (FILE_err&4) return (1);
  *y = (FSTAR_INT)FILE2_read_int (fp);
  if (FILE_err&4) return (1);
  if ( F->flag & LOAD_ID1 ){ x--; y--; }
  FILE2_read_until_newline ( fp);
  if ( F->edge_dir==-1 || (F->edge_dir == 0 && *x > *y) ) SWAP_FSTAR_INT (*x, *y);
  return (0);
}

/* compute node_num's and allocate arrays for degree */
void FSTAR_alloc_deg (FSTAR *F){
  F->out_node_num = F->xmax;
  F->in_node_num = F->ymax;
  F->node_num = MAX (F->out_node_num, F->in_node_num);
  if ( F->edge_dir == -1 ) SWAP_FSTAR_INT ( F->out_node_num, F->in_node_num);
  if ( F->edge_dir == 0 ) F->in_node_num = F->out_node_num = F->node_num;
  calloc2 (F->fstar, F->out_node_num+1, EXIT);
  if ( F->flag & LOAD_EDGE ) return;

  if ( F->flag & FSTAR_CNT_IN ){
    if ( F->flag & FSTAR_CNT_DEG_ONLY ){
        calloc2 (F->in_deg, F->node_num+1, EXIT);
    } else calloc2 (F->in_deg, F->in_node_num+1, EXIT);
  }
  if ( F->flag & FSTAR_CNT_OUT )
      calloc2 (F->out_deg, F->out_node_num+1, EXIT);
}

/* compute fstar from outdegree stored in fstar itself, and allocate edge array */
void FSTAR_calc_fstar (FSTAR *F){
  FSTAR_INT i, j=0, jj;
  FLOOP (i, 0, F->out_node_num){
    jj = j + F->fstar[i];
    F->fstar[i] = j;
    j = jj;
  }
  F->fstar[i] = F->edge_num = j;
//  malloc2 (F->edge, ((F->flag&LOAD_EDGE)?2:1) *F->edge_num +2, EXIT);
//  ARY_FILL (F->edge, 0, ((F->flag&LOAD_EDGE)?2:1) *F->edge_num +2, F->node_num+1);
  malloc2 (F->edge, F->edge_num +2, EXIT);
  ARY_FILL (F->edge, 0, F->edge_num +2, F->node_num+1);
}

/* return 1 if edge (x,y) is valid */
/* head != 0 means x is the origin, otherwise x is destination */
int FSTAR_eval_edge (FSTAR *F, FSTAR_INT x, FSTAR_INT y){

  if ( x>F->node_num || y>F->node_num) return (0);
  if ( F->flag & LOAD_EDGE ){
    if ( F->fstar[F->out_node_num] ){
      if ( !RANGE (F->deg_lb, F->fstar[x+1]-F->fstar[x], F->deg_ub) ) return (0);
      if ( !RANGE (F->deg_lb, F->fstar[y+1]-F->fstar[y], F->deg_ub) ) return (0);
    } else {
      if ( !RANGE (F->deg_lb, F->fstar[x], F->deg_ub) ) return (0);
      if ( !RANGE (F->deg_lb, F->fstar[y], F->deg_ub) ) return (0);
    }
    return (1);
  }

  if ( F->out_deg && (F->flag&FSTAR_OUT_CHK)){
    if ( !RANGE (F->out_lb, F->out_deg[x], F->out_ub) ) return (0);
    if ( (F->flag&LOAD_BIPARTITE)==0
          && !RANGE(F->out_lb, F->out_deg[y], F->out_ub) ) return (0);
  }
  if ( F->in_deg && F->out_deg && (F->flag&LOAD_BIPARTITE)==0 && (F->flag&FSTAR_DEG_CHK)){
    if ( !RANGE (F->deg_lb, F->in_deg[x]+F->out_deg[x], F->deg_ub) ) return (0);
    if ( !RANGE (F->deg_lb, F->in_deg[y]+F->out_deg[y], F->deg_ub) ) return (0);
  }
  
  if ( ((F->flag & FSTAR_CNT_DEG_ONLY) || F->edge_dir==0) && F->in_deg && (F->flag&FSTAR_DEG_CHK)){
    if ( !RANGE (F->deg_lb, F->in_deg[x], F->deg_ub) ) return (0);
    if ( !RANGE (F->deg_lb, F->in_deg[y], F->deg_ub) ) return (0);
  } else if ( F->in_deg && (F->flag&FSTAR_IN_CHK) && !(F->flag&FSTAR_CNT_DEG_ONLY) ){
    if ( (F->flag&LOAD_BIPARTITE)==0
         && !RANGE (F->in_lb, F->in_deg[x], F->in_ub) ) return (0);
    if ( !RANGE (F->in_lb, F->in_deg[y], F->in_ub) ) return (0);
  }
  return (1);
}

/* scan the file and count the degree, for edge listed file */
void FSTAR_scan_file (FSTAR *F, FILE2 *fp){
  FSTAR_INT x, y, i;
  FILE_COUNT C;

    // count #pairs
  C = FILE2_count (fp, F->flag&(LOAD_ELE+LOAD_TPOSE+LOAD_NUM+LOAD_GRAPHNUM), 0, 0, 0, 0, 0);
  F->xmax = C.rows; F->ymax = C.clms; F->edge_num_org = C.eles;
  FSTAR_alloc_deg (F);

    // read pairs from file
  FILE2_reset (fp);
  if ( F->flag&(LOAD_NUM+LOAD_GRAPHNUM) ) FILE2_read_until_newline (fp);
  
  if ( F->flag & LOAD_ELE ){
    FLOOP (i, 0, F->edge_num_org){
      while ( FSTAR_read_pair (F, fp, &x, &y));
      F->fstar[x]++;
      if ( F->flag & LOAD_EDGE ) F->fstar[y]++;
      FSTAR_inc_deg (F, x, y);
    }
  } else {
    i=0;
    do {   // compute xmax and ymax
      do {  // loop for read one line
        y = (FSTAR_INT)FILE2_read_int (fp);
        if ( F->flag & LOAD_ID1 ) y--;
        if ( FILE_err&4 ) break;
        x = i;
        if ( F->edge_dir==-1 || (F->edge_dir == 0 && x > y) ) SWAP_FSTAR_INT (x, y);
        F->fstar[x]++;
        if ( F->flag & LOAD_EDGE ) F->fstar[y]++;
        FSTAR_inc_deg (F, x, y);
      } while ( (FILE_err&3)==0 );
      i++;
    } while ( (FILE_err&2)==0 );
  }
}

/* read induced subgraph by node_mark; need node_deg, node_mark and node_num */
/* node_deg will be accumerated */
void FSTAR_read_file_edge (FSTAR *F, FILE2 *fp){
  FSTAR_INT i, x, y;
  
  FILE2_reset (fp);
  if ( F->flag&(LOAD_NUM+LOAD_GRAPHNUM) ) FILE2_read_until_newline (fp);

  if ( F->flag & (FSTAR_DEG_CHK+FSTAR_IN_CHK+FSTAR_OUT_CHK) ){
    FLOOP (i, 0, F->edge_num_org){
      while ( FSTAR_read_pair (F, fp, &x, &y) );
      if ( FSTAR_eval_edge (F, x, y) == 0 ){
        F->fstar[x]--;
        if ( F->flag & LOAD_EDGE ) F->fstar[y]--;
      }
    }
  }

  FSTAR_calc_fstar (F);
  FILE2_reset (fp);
  if ( F->flag&(LOAD_NUM+LOAD_GRAPHNUM) ) FILE2_read_until_newline (fp);

  FLOOP (i, 0, F->edge_num_org){
    while ( FSTAR_read_pair (F, fp, &x, &y) );
    if ( F->fstar[x+1] > F->fstar[x] && F->fstar[y+1]>F->fstar[y] ){
       F->edge[F->fstar[x]++] = y;
      if ( F->flag & LOAD_EDGE ){ F->edge[F->fstar[y]++] = x; }
    }
  }
  BLOOP (i, F->out_node_num, 0) F->fstar[i+1] = F->fstar[i];
  F->fstar[0] = 0;
}

/* read induced subgraph by node_mark; need node_deg, node_mark and node_num */
/* node_deg will be accumerated */
void FSTAR_read_file_node (FSTAR *F, FILE2 *fp){
  FSTAR_INT i, x, y, j = (F->flag & (FSTAR_DEG_CHK+FSTAR_IN_CHK+FSTAR_OUT_CHK))? 2: 1;

  while (j>0){
    i=0;
    FILE2_reset (fp);
    if ( F->flag&(LOAD_NUM+LOAD_GRAPHNUM) ) FILE2_read_until_newline (fp);
    if ( j==1 ) FSTAR_calc_fstar (F);
    do {   // compute xmax and ymax
      do {  // loop for read one line
        y = (FSTAR_INT)FILE2_read_int (fp);
        if ( F->flag & LOAD_ID1 ) y--;
        if ( FILE_err&4 ) break;
        x = i;
        if ( F->edge_dir==-1 || (F->edge_dir == 0 && x > y) ) SWAP_FSTAR_INT (x, y);
        if ( FSTAR_eval_edge (F, x, y) ){
          if ( j==1 ){
            F->edge[F->fstar[x]++] = y;
            if ( F->flag & LOAD_EDGE ) F->edge[F->fstar[y]++] = x;
          }
        } else if ( j==2 ){
          F->fstar[x]--;
          if ( F->flag & LOAD_EDGE ) F->fstar[y]--;
        }
      } while ( !(FILE_err&3) );
      i++;
    } while ( !(FILE_err&2) );
    j--;
  }
  BLOOP (i, F->out_node_num, 0) F->fstar[i+1] = F->fstar[i];
  F->fstar[0] = 0;
}

/* remove edges adjacent to small/large degree vertices */
void FSTAR_extract_subgraph (FSTAR *F){
  FSTAR_INT x, y, ii, i, nodes;
  int flag = 0;
  
    // count #pairs
  do {
    nodes = 0;
    if ( F->in_deg || F->out_deg ){
        // re-count in/out degrees
      if ( F->in_deg ) ARY_FILL (F->in_deg, 0, F->in_node_num, 0);
      if ( F->out_deg ) ARY_FILL (F->out_deg, 0, F->out_node_num, 0);
      x=0; FLOOP (i, 0, F->edge_num){
        while ( i == F->fstar[x+1] ) x++;
        FSTAR_inc_deg (F, x, F->edge[i]);
      }
    }
      // re-remove out-bounded degree vertices
    ii=x=0; FLOOP (i, 0, F->edge_num){
      while ( i == F->fstar[x+1] ){ F->fstar[x+1] = ii; x++; flag = 0; }
      y = F->edge[i];
      if ( FSTAR_eval_edge (F, x, y) ){
        F->edge[ii++] = y;
        if ( flag==0 ){ nodes++; flag = 1; }
      }
    }
    F->edge_num = ii;
    print_mes (F, "iterative scan: #nodes="FSTAR_INTF", #edges = %d\n", nodes, i);
  } while ( ii < i );
}

/* load graph from file */
void FSTAR_load (FSTAR *F){
  FILE2 fp = INIT_FILE2;
  FSTAR_init2 (F);

  FILE2_open (fp, F->fname, "r", EXIT);
  FSTAR_scan_file (F, &fp);
  print_mes (F, "fisrst & second scan end: %s\n", F->fname);

  if ( F->flag & LOAD_ELE ) FSTAR_read_file_edge (F, &fp);
  else FSTAR_read_file_node (F, &fp);
  FILE2_close(&fp);   if (ERROR_MES) EXIT;
  print_mes (F, "file read end: %s\n", F->fname);

  FSTAR_extract_subgraph (F);
//  if ( ((F->flag&LOAD_INCSORT) && !(F->flag&LOAD_ELE))
  if ( (F->flag&LOAD_INCSORT)
     || (F->flag&LOAD_DECSORT) ) FSTAR_sort_adjacent_node (F);
  print_mes (F, "forwardstar graph: %s ,#nodes %d(%d,%d) ,#edges %d\n", F->fname, F->node_num, F->in_node_num, F->out_node_num, F->edge_num);
}

/* make vertex permutation table and write to table-file */
void FSTAR_write_table_file (FSTAR *F, char *fname){
  FSTAR_INT i, flag;
  FILE *fp;
  if ( !fname ) return;
  fopen2 (fp, fname, "w", EXIT);
  calloc2 (F->table, F->node_num, {fclose(fp);EXIT;});
  F->reduced_node_num = 0;
  FLOOP (i, 0, F->out_node_num){
    flag = 0;
    if ( F->out_deg ) flag += F->out_deg[i];
    if ( F->in_deg && (F->in_node_num==0 || i < F->in_node_num) ) flag += F->in_deg[i];
    if ( flag ){
      fprintf (fp, FSTAR_INTF"\n", (F->flag&LOAD_ID1)?i+1:i);
      F->table[i] = F->reduced_node_num;
      F->reduced_node_num++;
    } else F->table[i] = F->out_node_num;
  }
  fclose2 (fp);
}

FILE *FSTAR_open_write_file (FSTAR *F, char *fname){
  FILE *fp;
  fopen2 (fp, fname, "w", EXIT);
  if ( F->flag&(LOAD_NUM+LOAD_GRAPHNUM) ){
    fprintf (fp, FSTAR_INTF" ", F->node_num);
    if ( F->flag&LOAD_NUM ) fprintf (fp, FSTAR_INTF" ", F->node_num);
    fprintf (fp, FSTAR_INTF"\n", F->edge_num/((F->flag&LOAD_EDGE)?2:1));
  }
  return (fp);
}

/* write induced graph and the number convert table; needs accumerated node_deg, edge */
void FSTAR_write_graph_edge (FSTAR *F, char *fname){
  FILE *fp = FSTAR_open_write_file (F, fname);
  FSTAR_INT i, x=0, xx, y;
  
  FLOOP (i, 0, F->edge_num){
    while ( i == F->fstar[x+1] ) x++;
    y = F->edge[i];
    if ( F->table ){
      xx = F->table[x];
      if ( !(F->flag&LOAD_BIPARTITE)) y = F->table[y];
    } else xx = x;
    if ( F->flag & LOAD_ID1 ) y++;
    fprintf (fp, FSTAR_INTF" "FSTAR_INTF"\n", xx, y);
  }
  fclose (fp);
}

/* write induced graph and the number convert table; needs accumerated node_deg, edge */
void FSTAR_write_graph_node (FSTAR *F, char *fname){
  FILE *fp = FSTAR_open_write_file (F, fname);
  FSTAR_INT i, x=0, y, flag, c=0, row=0;

  FLOOP (i, 0, F->edge_num+1){
    while ( i == F->fstar[x+1]){
      flag = 0;
      if ( F->out_deg && (F->out_node_num ==0 || x < F->out_node_num) ) flag += F->out_deg[x];
      if ( F->in_deg && (F->in_node_num ==0 || x < F->in_node_num) ) flag += F->in_deg[x];
      if ( (!F->out_deg && !F->in_deg) && (F->flag&LOAD_EDGE) ) flag = F->fstar[x+1] - F->fstar[x];
      if ( !F->table || flag ){ fprintf (fp, "\n"); c++; row=0; }
      if ( ++x == F->out_node_num ) goto END;
    }
    y = F->edge[i];
    if ( !(F->flag&(LOAD_BIPARTITE+LOAD_EDGE)) && F->edge_dir==0 && x>y ) continue;
    if ( F->table && !(F->flag&LOAD_BIPARTITE) ){
      if ( (y=F->table[y]) == F->out_node_num ) continue;
    }
    if ( F->flag & LOAD_ID1 ) y++;
    if (row==0 && (F->flag & FSTAR_INS_ROWID)){
      fprintf (fp, "%d ", c);
      row = 1;
    }
    fprintf (fp, FSTAR_INTF" ", y);
  }
  fprintf (fp, "\n");
  END:;
  fclose ( fp);
}

#endif

